#include <algorithm>
#include "MusicQueue.h"
#include "MusicClient.h"

using namespace ts;
using namespace ts::server;
using namespace ts::music;
using namespace ::music;
using namespace std;
using namespace std::chrono;

MusicQueue::MusicQueue(ts::server::MusicClient* self) {
	this->handle = self;
}

MusicQueue::~MusicQueue() {}

std::shared_ptr<PlayableSong> MusicQueue::currentSong() {
	return this->current;
}

void MusicQueue::notifySongEnded() {
	if(this->currentSong()) {
		this->getNextSong();
	}
}

void MusicQueue::notifySongAbort() {
	this->getNextSong(); //TODO retry count!
}

std::shared_ptr<PlayableSong> MusicQueue::rewind() {
	threads::MutexLock lock(this->lock);

	if(this->current)
		this->queue.push_front(std::move(this->current));
	if(!this->_history.empty()) {
		this->current = dynamic_pointer_cast<PlayableSong>(this->_history.front());
		this->_history.pop_front();
	}

	return this->current;
}

std::shared_ptr<PlayableSong> MusicQueue::getNextSong() {
	threads::MutexLock lock(this->lock);
	if(this->current) {
		this->_history.push_front(this->current);
		for(int index = 10; index < this->_history.size(); index++) { //Last ten songs full cache!
			if(!dynamic_pointer_cast<PlayableSong>(this->_history[index])) break;

			auto element = this->_history[index];
			this->_history[index] = make_shared<MusicHistoryEntry>(element->getUrl(), element->getInvoker(), element->getSongId());
		}
	}

	if(this->queue.empty()) this->current = nullptr;
	else {
		this->current = this->queue.front();
		this->queue.pop_front();
	}
	return this->current;
}

std::shared_ptr<PlayableSong> MusicQueue::find_queue(ts::SongId id) {
	threads::MutexLock lock(this->lock);
	for(const auto& element : this->queue)
		if(element->getSongId() == id)
			return element;
	return nullptr;
}

std::shared_ptr<SongInfo> MusicQueue::find(ts::SongId id) {
	threads::MutexLock lock(this->lock);
	for(const auto& element : this->queue)
		if(element->getSongId() == id)
			return element;
	for(const auto& element : this->_history)
		if(element->getSongId() == id)
			return element;
	if(this->current && this->current->getSongId() == id)
		return this->current;

	return nullptr;
}

std::deque<std::shared_ptr<PlayableSong>> MusicQueue::queueEntries() {
	return this->queue;
}

std::deque<std::shared_ptr<SongInfo>> MusicQueue::history() {
	return this->_history;
}

bool MusicQueue::deleteEntry(const std::shared_ptr<PlayableSong>& entry) {
	if(!entry) return false;

	threads::MutexLock lock(this->lock);
	auto found = std::find(this->queue.begin(), this->queue.end(), entry);
	if(found == this->queue.end()) return false;
	this->queue.erase(found);
	return true;
}

int MusicQueue::changeOrder(const std::shared_ptr<PlayableSong>& entry, int position) {
	threads::MutexLock lock(this->lock);
	auto found = std::find(this->queue.begin(), this->queue.end(), entry);
	if(found == this->queue.end()) return -1;
	this->queue.erase(found);

	if(position < 0) position = (int) this->queue.size();
	if(position > this->queue.size()) position = (int) this->queue.size();
	this->queue.insert(this->queue.begin(), entry);
	return position;
}

std::shared_ptr<PlayableSong> MusicQueue::insertEntry(const std::string &url, const tsclient &invoker, const std::shared_ptr<PlayableSong::song_loader_t> &loader) {
	auto entry = PlayableSong::create({
		0,
		url,
		invoker ? invoker->getClientDatabaseId() : 0
	}, loader);
	{
		entry->set_song_id(this->song_id_index++);
		threads::MutexLock lock(this->lock);
		this->queue.push_back(entry);
	}

	entry->get_loader(this->handle->getServer(), true);
	return entry;
}