#include <dlfcn.h>
#include "ChannelProvider.h"
#include "../../MusicClient.h"
#include "../../../../InstanceHandler.h"
#include "../../../../server/file/FileServer.h"
#include "../../../../channel/ServerChannel.h"
#include "../../../../../../music/providers/ffmpeg/FFMpegProvider.h"


using namespace std;
using namespace music::provider;
using namespace music::manager;
using namespace music;
using namespace ts;
using namespace ts::server;

struct AudioFileInfo {
	std::string absolute_path;
	std::string description;
};

extern ts::server::InstanceHandler* serverInstance;
threads::Future<std::shared_ptr<::music::MusicPlayer>> ChannelProvider::createPlayer(const std::string &url, void*, void* ptr_server) {
	auto server = ((TSServer*) ptr_server)->ref();
	threads::Future<std::shared_ptr<::music::MusicPlayer>> future;

	if(server) {
		std::thread([future, server, url, ptr_server]{
			auto f_server = serverInstance->getFileServer();
			if(!f_server) {
				future.executionFailed("instance missing file server");
				return;
			}

			Command command("");
			try {
				command = Command::parse(pipes::buffer_view{url.data(), url.length()}, false);
			} catch(std::exception& ex) {
				future.executionFailed("failed to parse data");
				return;
			}

			try {
				auto channel_id = command["cid"].as<ChannelId>();
				auto name = command["name"].string();
				auto path = command["path"].string();

				auto channel = server->getChannelTree()->findChannel(channel_id);
				if(!channel) {
					future.executionFailed("could not find target channel");
					return;
				}

				auto directory = f_server->resolveDirectory(server, channel, path);
				if(!directory) {
					future.executionFailed("invalid file path");
					return;
				}

				auto file = f_server->findFile(name, directory);
				if(!file) {
					future.executionFailed("could not find file");
					return;
				}

				if(file->type != file::FileType::FILE) {
					future.executionFailed("file isnt a file");
					return;
				}

				auto info = make_shared<AudioFileInfo>();
				info->absolute_path = file->path + "/" + file->name;
				info->description = file->name;

				auto ffmpeg = ::music::manager::resolveProvider("FFMpeg", "");
				if(!ffmpeg) {
					future.executionFailed("missing ffmpeg");
					return;
				}

				{
					auto ffmpeg_data = (::music::FFMpegData::FileReplay*) malloc(sizeof(::music::FFMpegData::FileReplay));
					if(!ffmpeg_data) {
						future.executionFailed("failed to allocate memory");
						return;
					}

					ffmpeg_data->version = ::music::FFMpegData::CURRENT_VERSION;
					ffmpeg_data->type = ::music::FFMpegData::REPLAY_FILE;
					ffmpeg_data->_free = ::free;

					ffmpeg_data->file_path = (char*) malloc(info->absolute_path.length() + 1);
					ffmpeg_data->file_path[info->absolute_path.length()] = '\0';
					ffmpeg_data->file_description = (char*) malloc(info->description.length() + 1);
					ffmpeg_data->file_description[info->description.length()] = '\0';

					memcpy(ffmpeg_data->file_path, info->absolute_path.data(), info->absolute_path.length());
					memcpy(ffmpeg_data->file_description, info->description.data(), info->description.length());

					auto p = ffmpeg->createPlayer("", ffmpeg_data, ptr_server);
					p.wait();
					if(p.failed())
						future.executionFailed(p.errorMegssage());
					else
						future.executionSucceed(*p.get());
				}
				return;
			} catch(std::exception& ex) {
				future.executionFailed("failed to process data");
				return;
			}
		}).detach();
	} else {
		future.executionFailed("invalid bot");
	}


	return future;
}

bool ChannelProvider::acceptString(const std::string &str) {
	if(str.find("cid=") == string::npos) return false;
	if(str.find("path=") == string::npos) return false;
	if(str.find("name=") == string::npos) return false;

	return true;
}

vector<string> ChannelProvider::availableFormats() {
	return {};
}

vector<string> ChannelProvider::availableProtocols() {
	return {"query"};
}

std::shared_ptr<::music::manager::PlayerProvider> ChannelProvider::create_provider() {
	/*
	std::shared_ptr<yt::YTProviderConfig> config = make_shared<yt::YTProviderConfig>();

	{
		auto config_path = fs::u8path("providers/config_channel.ini");
		music::log::log(music::log::debug, "[YT-DL] Using config file located at " + config_path.string());
		if(fs::exists(config_path)) {
			INIReader ini_reader(config_path.string());

			if(ini_reader.ParseError()) {
				music::log::log(music::log::err, "[YT-DL] Could not parse config! Using default values");
			} else {
				config->youtubedl_command = ini_reader.Get("general", "youtubedl_command", config->youtubedl_command);
				config->commands.version = ini_reader.Get("commands", "version", config->commands.version);
				config->commands.query_video = ini_reader.Get("commands", "query_video", config->commands.query_video);
				music::log::log(music::log::info, "[YT-DL] Config successfully loaded");
			}
		} else {
			music::log::log(music::log::debug, "[YT-DL] Missing configuration file. Using default values");
		}
	}
	*/

	return std::shared_ptr<ChannelProvider>(new ChannelProvider(), [](ChannelProvider* provider){
		if(!provider) return;
		delete provider;
	});
}

threads::Future<shared_ptr<UrlInfo>> ChannelProvider::query_info(const std::string &string, void *pVoid, void *pVoid1) {
	auto info = make_shared<UrlSongInfo>();

	info->type = UrlType::TYPE_VIDEO;
	info->url = string;
	info->title = "";
	info->description = "";
	info->metadata = {};

	threads::Future<shared_ptr<UrlInfo>> result;
	result.executionSucceed(info);
	return result;
}
