//
// Created by wolverindev on 07.12.17.
//

#include "ApplicationInteraction.h"
#include <log/LogUtils.h>
#include <sys/shm.h>
#include <cstring>
#include <iostream>
#include <zconf.h>

using namespace std;

static key_t SHM_KEY = 0x1337015; //1337 TS
static size_t SHARED_SIZE = sizeof(interaction::DefaultMemoryInfo);

static int SHARED_MEMORY_ID = -1;
static void* SHARED_MEMORY_PTR = nullptr;
static size_t INSTANCE_ID = 0;

namespace interaction {
    bool waitForAttach(std::string& error){
        SHARED_MEMORY_ID = shmget(SHM_KEY, SHARED_SIZE, IPC_CREAT | 0776);
        if(SHARED_MEMORY_ID < 0) {
            error = string() + "Could not attach: " + strerror(errno);
            return false;
        }

        shmid_ds shmInfo{};
        memset(&shmInfo, 0, sizeof(shmInfo));
        if(shmctl(SHARED_MEMORY_ID, IPC_STAT, &shmInfo) < 0){
            error = string() + "Could not collect info: " + strerror(errno);
            return false;
        }

        SHARED_MEMORY_PTR = shmat(SHARED_MEMORY_ID, nullptr, 0);
        if(SHARED_MEMORY_PTR == (char*) - 1){
            SHARED_MEMORY_PTR = nullptr;
            error = "failed to attach";
            return false;
        }

        DefaultMemoryInfo* info = nullptr;
        if(shmInfo.shm_nattch == 0){ //Im the first program!
            debugMessage("Created new memory struct");
            *(DefaultMemoryInfo*) SHARED_MEMORY_PTR = DefaultMemoryInfo{};
            info = (DefaultMemoryInfo*) SHARED_MEMORY_PTR;

            pthread_mutex_lock(&info->memoryLock);
            memset(info->infoUsed, 0, sizeof(info->infoUsed) / sizeof(info->infoUsed[0]));
        } else {
            info = memoryInfo();
            pthread_mutex_lock(&info->memoryLock);
        }

        if(info->instanceCount + 1 >= info->maxInstances){
            pthread_mutex_unlock(&info->memoryLock);
            error = "to many instances";
            return false;
        }

        for(int index = 0; index < info->maxInstances; index++)
            if(!info->infoUsed[index]){
                info->instanceCount++;
                INSTANCE_ID = static_cast<size_t>(index);
                break;
            }

        info->info[INSTANCE_ID] = InstanceInfo{};
        info->infoUsed[INSTANCE_ID] = true;

        auto instanceInfo = ownInfo();
        instanceInfo->pid = ::getpid();
        pthread_mutex_unlock(&info->memoryLock);
        debugMessage("Got application instance id " + to_string(INSTANCE_ID));
        return true;
    }

    void removeMemoryHook(){
        if(SHARED_MEMORY_PTR){
            if(shmdt(SHARED_MEMORY_PTR) < 0)
                cerr << strerror(errno) << endl;
            SHARED_MEMORY_PTR = nullptr;
        }

        shmid_ds info{};
        memset(&info, 0, sizeof(info));
        if(shmctl(SHARED_MEMORY_ID, IPC_STAT, &info) < 0){
            logMessage(lstream << "Could not collect info: " << strerror(errno));
            return;
        }

        if(info.shm_nattch == 0){
            shmctl(SHARED_MEMORY_ID, IPC_RMID, nullptr);
            logMessage("Deleting shared memory");
        }
    }

    DefaultMemoryInfo* memoryInfo(){ return static_cast<DefaultMemoryInfo *>(SHARED_MEMORY_PTR); }
    InstanceInfo* ownInfo(){ return &memoryInfo()->info[INSTANCE_ID]; }
}