#pragma once

#include <vector>
#include <string>
#include <chrono>
#include <memory>
#include <ThreadPool/Thread.h>
#include <deque>
#include <ThreadPool/Mutex.h>
#include <netinet/in.h>
#include <sqlite3.h>
#include <sql/SqlQuery.h>
#include "../Group.h"
#include <event.h>
#include "../manager/IpListManager.h"

namespace ts {
    class BasicChannel;
    class Group;

    namespace server {
        class TSServer;
        class QueryClient;

        struct QueryLoginCredentials {
            std::string username;
            std::string password;
            ClientUid uniqueId;
        };

	    class QueryAccount {
	    	public:
				std::string username;
		        std::string unique_id;

		        ServerId bound_server;

			    QueryAccount() = default;
			    virtual ~QueryAccount() = default;
        };

	    class PasswortedQueryAccount : public QueryAccount {
		    public:
				std::string password;

			    PasswortedQueryAccount() = default;
			    virtual ~PasswortedQueryAccount() = default;
        };

        class QueryServer {
                friend class QueryClient;
            public:
                explicit QueryServer(sql::SqlManager*);
                ~QueryServer();

                bool start(const sockaddr_in&, std::string&);
                void stop();
                bool running(){ return active; }
                sockaddr_in* boundedAddress(){ return boundAddress; }

                void unregisterConnection(const std::shared_ptr<QueryClient> &);

                /*
                std::string createQueryLogin(const std::string &name, ClientUid uid, std::string = "");
                bool renameQueryLogin(ClientUid uid, const std::string &targetName);
		        std::string resetQueryPassword(const  std::shared_ptr<QueryLoginCredentials>&, std::string = "");

                std::shared_ptr<QueryLoginCredentials> findQueryLoginByName(const std::string &name);
                std::shared_ptr<QueryLoginCredentials> findQueryLoginByUid(const std::string &uid);
				*/

		        std::deque<std::shared_ptr<QueryAccount>> list_query_accounts(OptionalServerId /* server */);
		        std::shared_ptr<QueryAccount> create_query_account(const std::string& /* name */, ServerId /* server */, const std::string& /* owner unique id */, const std::string& /* password */);
		        std::shared_ptr<PasswortedQueryAccount> load_password(const std::shared_ptr<QueryAccount>& /* account */);
		        bool delete_query_account(const std::shared_ptr<QueryAccount>& /* account */);

		        std::shared_ptr<QueryAccount> find_query_account_by_name(const std::string& /* name */);
		        std::deque<std::shared_ptr<QueryAccount>> find_query_accounts_by_unique_id(const std::string &unique_id /* unique id */);

		        bool rename_query_account(const std::shared_ptr<QueryAccount>& /* account */, const std::string& /* new name */);
		        bool change_query_password(const std::shared_ptr<QueryAccount>& /* account */, const std::string& /* new password */);

		        threads::ThreadPool* executePool() { return this->_executePool; }
            private:
                sql::SqlManager* sql;
                sockaddr_in* boundAddress = nullptr;

                bool active = false;
                int serverSocket;
                std::vector<threads::Thread*> threads;

                std::unique_ptr<IpListManager> ip_whitelist;
		        std::unique_ptr<IpListManager> ip_blacklist;

                //IO stuff
                event_base* eventLoop = nullptr;
                ::event* acceptEvent = nullptr;

                threads::ThreadPool* _executePool = nullptr;

                threads::Mutex clientLock;
                std::deque<std::shared_ptr<QueryClient>> connectedClients;

                threads::Mutex loginLock;
                std::chrono::system_clock::time_point lastDecrease;
                std::map<std::string, uint32_t> loginAttempts;
                std::map<std::string, std::chrono::system_clock::time_point> queryBann;

                threads::Thread* ioThread = nullptr;
                threads::SchedulingTask tickingId = 0;
                void onClientAccept(int fd, short ev, void *arg);
                void tick();
        };
    }
}