#pragma once

#include <ThreadPool/Mutex.h>
#include <ThreadPool/Thread.h>
#include <event.h>

namespace ts {
    namespace server {
        class TSServer;
        class WebClient;
        class WebRTCServer;

        class WebControlServer {
                friend class WebClient;
            public:
                explicit WebControlServer(const std::shared_ptr<TSServer>&);
                ~WebControlServer();

                bool start(sockaddr_in*, std::string&);
                inline bool running(){ return _running; }
                void stop();

                std::shared_ptr<TSServer> getTS(){ return this->handle; }

                std::deque<std::shared_ptr<WebClient>> connectedClients(){
                    threads::MutexLock l(this->clientLock);
                    return this->clients;
                }

            private:
                std::shared_ptr<TSServer> handle = nullptr;

                threads::Mutex clientLock;
                std::deque<std::shared_ptr<WebClient>> clients;

                bool _running = false;
                sockaddr_in* boundAddress = nullptr;
                int serverSocket = 0;
                ::event* acceptEvent = nullptr;
            private:
                void onClientAccept(int fd, short ev, void *arg);
                void unregisterConnection(const std::shared_ptr<WebClient>&);
        };
    }
}